import type { vec3, vec4 } from '../constants.js';
/**
 * *Common utilities for working with colors in vec3, vec4, or hexadecimal form.*
 *
 * Provides methods to convert linear components (vec3, vec4) to sRGB hex values. All colors in
 * the glTF specification, excluding color textures, are linear. Hexadecimal values, in sRGB
 * colorspace, are accessible through helper functions in the API as a convenience.
 *
 * ```typescript
 * // Hex (sRGB) to factor (linear).
 * const factor = ColorUtils.hexToFactor(0xFFCCCC, []);
 *
 * // Factor (linear) to hex (sRGB).
 * const hex = ColorUtils.factorToHex([1, .25, .25])
 * ```
 *
 * @category Utilities
 */
export declare class ColorUtils {
    /**
     * Converts sRGB hexadecimal to linear components.
     * @typeParam T vec3 or vec4 linear components.
     */
    static hexToFactor<T = vec3 | vec4>(hex: number, target: T): T;
    /**
     * Converts linear components to sRGB hexadecimal.
     * @typeParam T vec3 or vec4 linear components.
     */
    static factorToHex<T = vec3 | vec4>(factor: T): number;
    /**
     * Converts sRGB components to linear components.
     * @typeParam T vec3 or vec4 linear components.
     */
    static convertSRGBToLinear<T = vec3 | vec4>(source: T, target: T): T;
    /**
     * Converts linear components to sRGB components.
     * @typeParam T vec3 or vec4 linear components.
     */
    static convertLinearToSRGB<T = vec3 | vec4>(source: T, target: T): T;
}
